#ifdef	FAKE_UNICODE

/*
 * detect_unicode_stream(): call to check either input text is UNICODE or not.
 *
 * Returns:
 *  0 - ANSI text or binary file;
 *  1 - UTF8 text;
 *  2 - UNICODE text, little-endian;
 *  3 - UNICODE text, big-endian.
 * Side effects: if unicode text detected, file pointer moved to the
 * start of data.
 */
V_LOCAL int detect_unicode_stream(file_t in_stream)
{
	char head[3];
	fileRead(in_stream,head,2);
	if(head[0] == 0xFF && head[1] == 0xFE)
		return 2; /* LE UNICODE */
	if(head[0] == 0xFE && head[1] == 0xFF)
		return 3; /* BE UNICODE */
	fileRead(in_stream,head+2,1);
	if(head[0] == 0xEF && head[1] == 0xBB && head[2] == 0xBF)
		return 1; /* UTF8 */
	fileSeek(in_stream,-3,SEEK_CUR);
	return 0;
}

u8_t utf_to_cyrillic[] =
{
  /* Ũ */
  0xD0, 0x90, 0xD0, 0x91, 0xD0, 0x92,
  0xD0, 0x93, 0xD0, 0x94, 0xD0, 0x95,
  0xD0, 0x81, 0xD0, 0x96, 0xD0, 0x97, 
  0xD0, 0x98, 0xD0, 0x99, 0xD0, 0x9A, 
  0xD0, 0x9B, 0xD0, 0x9C, 0xD0, 0x9D, 
  0xD0, 0x9E, 0xD0, 0x9F, 0xD0, 0xA0, 
  0xD0, 0xA1, 0xD0, 0xA2, 0xD0, 0xA3, 
  0xD0, 0xA4, 0xD0, 0xA5, 0xD0, 0xA6, 
  0xD0, 0xA7, 0xD0, 0xA8, 0xD0, 0xA9, 
  0xD0, 0xAC, 0xD0, 0xAB, 0xD0, 0xAA, 
  0xD0, 0xAD, 0xD0, 0xAE, 0xD0, 0xAF,
  /*  */
  0xD0, 0xB0, 0xD0, 0xB1, 0xD0, 0xB2,
  0xD0, 0xB3, 0xD0, 0xB4, 0xD0, 0xB5,
  0xD1, 0x91, 0xD0, 0xB6, 0xD0, 0xB7,
  0xD0, 0xB8, 0xD0, 0xB9, 0xD0, 0xBA,
  0xD0, 0xBB, 0xD0, 0xBC, 0xD0, 0xBD,
  0xD0, 0xBE, 0xD0, 0xBF, 0xD1, 0x80,
  0xD1, 0x81, 0xD1, 0x82, 0xD1, 0x83,
  0xD1, 0x84, 0xD1, 0x85, 0xD1, 0x86,
  0xD1, 0x87, 0xD1, 0x88, 0xD1, 0x89,
  0xD1, 0x8C, 0xD1, 0x8B, 0xD1, 0x8A,
  0xD1, 0x8D, 0xD1, 0x8E, 0xD1, 0x8F
}; 

V_LOCAL u8_t get_utf8_char(vhandle_t in_stream)
{
	u8_t ch, ch2;
	fileRead(in_stream,&ch,1);
	if(ch < 128)
		return ch;
	fileRead(in_stream,&ch2,1);
	/* HACK */
	if(ch != 0xD0 && ch != 0xD1)
		return ch2;
	else
		for(i=0;i<132;i+=2)
			if(utf_to_cyrillic[i+1]==ch2 && utf_to_cyrillic[i]==ch)
				break;
	i <<= 1;
	if(i < 33)
	{
		/* Uppercase cyrillic letters; FIXME */
		return "Ũ"[i];
	} else if(i < 66)
	{
		/* Lowercase cyrillic letters; FIXME */
		i -= 33;
		return ""[i];
	} else
		return ch2;
}

#endif	/* FAKE_UNICODE */
